/*
** $Header: qpacket.h,v 2.1.1.3 88/07/11 11:18:29 donovan Exp $
**
** MODULE:  ~gamma/src/include3.1/qpacket.h
**
** This include file describes the structure of the
** query packets which are sent by query managers
** and exchanged by schedulers and query tasks.
**
** A more complete discussion of the meaning and
** use of these structures is contained in the document:
** ~gamma/doc/qpacket.nr 
**
*/

typedef enum OPERAND_TYPE {DUMMY1, NOCONST, STRINGCONST, INTCONST, FLOATCONST }; 

typedef enum QPKT_TYPE 	{DUMMY2, QUERY_PKT, QTNODE_PKT , CONTROL_PKT, DATA_PKT };

typedef enum SPLIT_TYPE {DUMMY3, HASHED_SPLIT, RANGE_SPLIT, RANDOM_SPLIT, NO_SPLIT,
				REHASH_SPLIT, AGGR_SPLIT };

typedef enum SCAN_TYPE 	{DUMMY4, SEGMENT_SCAN, INDEX_SCAN, NET_SCAN};

typedef enum INDEX_TYPE {DUMMY5, NO_INDEX, HASH_INDEX, BTREE_INDEX };

typedef enum PART_TYPE {DUMMY6, PIPELINED, STORED, OVERFLOW };

typedef enum OPCODE 	{DUMMY7, SELECT, AGGR_SCH, AGGR, AGGR_R, AGGRF_L, 
	AGGRF_R, SDELETE, CDELETE, APPEND, SREPLACE, CREPLACE, INSRT,
      JOIN , DUPL_ELIM , CREATE_REL, DESTROY_REL, CREATE_INDEX, 
      DESTROY_INDEX , STORE, DELAY_OP, SPOOL, MERGE, SORT, 
      TC_DELAY, TC_JOIN };

#define QTNODE_DESC(qpkt) ((char *)qpkt + sizeof(QPKT_HEADER))
#define QTNODE_OP(qtnode) ((char *)qtnode + sizeof(QTNODE_HEADER))
#define HASH_OPERATOR(opcode) ((opcode==JOIN) ||(opcode==DUPL_ELIM)\
				||(opcode==TC_JOIN))

/* define the number of lost bytes for host to node communication because
** of required message headers for nci.
**
** NOTE:  If you change this constant (and this should only change if nci
**	headers change) you MUST, MUST, MUST update the constant PROTO_LOSS
**	in file netpage.h  AND change MAXPACKET below!!!!!!!!
**	ALSO:  update constant STARTQUERY in file gdel_lib.h  and DTHEADERSIZE
**	in the scheduler deadlock directory in file defs.h.
**	These are also equal to constant DTHEADERSIZE below.
*/
#define UNIX_PROTEON_HEADER 36

#define PRINTOP(name,node)\
    if (node == NULLQTNODE)\
	printf ("%s = NIL\n",name);\
    else {\
	printf("%s = ",name);\
	print_opcode (node->opcode);\
    }

/* LOOK IN ~WISS/SRC/WISS.H (rel_op)
typedef	enum RELOP  { EQ, NE, LT, LE, GT, GE, GE_LE};
*/

/* default fillfactors and such */
#define DEFAULT_RELPAGES	50	/* pages arg to wiss create rel */
#define DEFAULT_RELEXTFF	99	/* extentfillfactor */
#define DEFAULT_PAGEFF		99	/* page fillfactor */

/* macros for computing references to descendent nodes */
#define LEFT(node) ((QTNODE_HEADER *)(context->packet + node->left))
#define RIGHT(node) ((QTNODE_HEADER *)(context->packet + node->right))

/* minimum required memory (kbytes) for a complex operation */
#define MIN_TBL_SIZE 64
/*#define	MAXHASH	65536*/
#define	MAXHASH	100000

/* maximum number of data bytes in a message */
#define MAXPACKET 8156 /* 8192 - 36 */

/* maximum number of attributes in a relation */
#define MAX_ATTR 50

/* one byte unsigned integer */
typedef unsigned char BYTE_INT;

/* space for the dt_header */
#define DTHEADERSIZE 36
typedef BYTE_INT DTHEADER[DTHEADERSIZE];

/* format of data portion of network packet under 3.1 */
/* commented out by dewitt in hypercube conversion
	typedef struct {
    	char	proto_bytes[DTHEADERSIZE];
    	char	data[2048 - DTHEADERSIZE];
	} PKT_BODY;
*/

/* 
** Query managers initialize PORT_TABLEs with values  
** that control which QTNODEs are activated at which
** sites.  The PRE_ACTIVE values of PORT_TABLES are
** replaced by valid port numbers when a query task is 
** activated.  At execution time, PORT_TABLEs represent
** a list of the network addresses of the processes
** that are cooperating on a particular query operation.
** Node names are implied by the corresponding positions
** of port values within the table. 
*/
#define NOT_ACTIVE	-1
#define PRE_ACTIVE	2
#define MAXNODE		32	
#define MAXDISK		32	
typedef BYTE_INT	PORT_TABLE[MAXNODE];

/* network address table */
typedef struct {
	BYTE_INT        node;
	BYTE_INT        port;
} NETADDR, NETTABLE[MAXNODE];


/* 
** NAME character arrays of this type contain attribute or relation names.
** STRINGVAL arrays are used to prefix/distinguish string attributes. 
*/
#define MAX_NAME 	32
typedef char 		NAME[MAX_NAME];

#define PREFIX	8
typedef char 		STRINGVAL[PREFIX];



/*
** This structure provides space for an attribute value of any type.
*/
typedef union {
	STRINGVAL	stringval;
	int		intval;
	float		floatval;
	double		doubleval;
} ANYVALUE;


/*
** Split table entries determine whether result tuples
** are sent to query pipes or are appended to stored relations.
** Split tables are stored at the beginning of the QTNODE_PARMS area.
*/
typedef struct {
	int	 	hashed_key;	/* upper limit of split range	*/
	ANYVALUE 	range_key;	/* intra-partition control key	*/
	int	 	overflow_key;	/* upper limit of split range	*/
	enum PART_TYPE	partition_type;	/* pipeline, stored or overflow	*/
	BYTE_INT	alt_hashfid;	/* alternate hash function name	*/
	BYTE_INT 	dest_node;	/* primary destination processor*/
	BYTE_INT 	dest_port;	/* primary destination port	*/
	BYTE_INT 	dest_stream;	/* buffered tuple stream	*/
	BYTE_INT	spool_address;	/* spooling processor		*/ 
	BYTE_INT	spool_port;	/* spooling port		*/
	BYTE_INT	spool_stream;	/* buffered tuple stream	*/
	char		*filter;	/* pointer to bit vector filter	*/
} SPLIT_ENTRY;


/*
** This structure describes the operational characteristics
** of attributes that are referenced by a query operator.
*/
typedef struct {
    	enum data_type	attr_type;	
	enum INDEX_TYPE	index_type; 	
	char		clustered;	/* 'C' or 'N'.  only set if */
					/* index_type is BTREE_INDEX */
	short 		position; 	/* ordinal position 		*/
	short 		offset;		/* bytes from start of tuple	*/
	short 		attrlen;	/* bytes in attribute 		*/
} ATTR_INFO;


/*
** This structure provides a representation for an operand 
** that may be either a constant or an attribute descriptor.
*/
typedef struct {
	enum OPERAND_TYPE operand_type;
	union {
		ATTR_INFO attribute;
		ANYVALUE constant;
	}info;
} OPERAND;



/*
** This structure defines clauses used by various operators.
** Sets of these clauses are stored in the QTNODE_PARMS component.
*/
typedef struct {
	OPERAND 	left;
	enum rel_op relop;
	OPERAND 	right;
	BYTE_INT 	drop_left ;
	BYTE_INT 	drop_right;
} CLAUSE;

/*  
**  This structure provides a representation for the information 
**  necessary to update all indices on a relation. The QT_NODE_PARAMS 
**  section of APPEND, SDELETE, CDELETE, SREPLACE, CREPACE packets 
**  contains an array of these structures.  The size of the array
**  is determined by the value of index_cnt in the 
**  QT_NODE_HDR of the packet.
*/

typedef struct {
	int		index_num;	/* index id */
	ATTR_INFO	index_attr;	/* attribute descriptor */
}  INDEX_DESC;


/*
** This descriptor contains the information required by the
** sdelete operator.
*/

typedef struct {
	NAME 		relname;	/* relation being scanned */
	int		tuple_length;	/* bytes in a tuple */
	int		index_cnt;	/* number of secondary indices to be 
					updated.  Index information is 
					contained in QT_NODE_PARAMS  */

	enum SCAN_TYPE	scan_type;	/* segment or index scan   */
	BYTE_INT 	clustered;
	/* if an index scan, specifies nature of index */
	INDEX_DESC 	scan_attr;	/* attribute used for index scan */
	int			range_operator[2];
	ANYVALUE 	range_value[2];	/* operands to range operator*/ 
	int		select_size;
	BYTE_INT	select_count;	/* # of clauses in QTNODE_PARMS	*/
} SDELETE_DESC;





/*
** This descriptor contains the information required by the
** scan operator (select/project).  
*/
typedef struct {
	NAME 		source_relname;	/* relation being scanned	*/
	int 		tuple_length;	/* bytes in a tuple		*/
	int		result_tuplen;	/* bytes in result tuple	*/
	int		index_num;	/* index id			*/
	enum SCAN_TYPE	scan_type;	/* segment or index scan	*/
	BYTE_INT 	clustered;
	/* if an index scan, specifies nature of index */
	ATTR_INFO 	scan_attr;	/* basis of the index		*/
	int range_operator[2];/* shrinks range of a scan	*/
	/* Assumes values 0 thru 6. It was originally of type enum_relop
	   but that clashed with some other IDM file */
	ANYVALUE 	range_value[2];	/* operands to range operator	*/ 
	BYTE_INT	rename;		/* boolean; rename + destroy	*/
	BYTE_INT	select_count;	/* # of clauses in QTNODE_PARMS	*/
	int		select_size;	/* # of bytes in boolean expr.	*/
	BYTE_INT	project_count;	/* # of attrs in QTNODE_PARMS	*/
	BYTE_INT	sort_input;	/* boolean; call wiss_sort	*/
	ATTR_INFO	sort_attr;
} SCAN_DESC;


/*
**  This is a descriptor for the append and insert operators
*/

typedef struct {
	NAME	relname;	/* name of relation to which tuple is to 
					be appended */
	int	tuple_length;	/* bytes in a tuple */
	int	index_cnt;	/* number of indices to be updated.  
				Index information is contained in 
				QT_NODE_PARAMS along with the actual tuple */
} APPEND_DESC;


/*
** This descriptor contains the information required by the
** sreplace operator.
*/

typedef struct {
	NAME 		relname;	/* relation being updated */
	int		tuple_length;	/* bytes in a tuple */
	int		index_cnt;	/* number of secondary indices to be 
					updated.  Index information is 
					contained in QT_NODE_PARAMS  */
	int		update_cnt;	/* number of attributes to be updated.
					Information on what attributes 
					to be updated is contained in
					QT_NODE_PARAMS.   */

	enum SCAN_TYPE	scan_type;	/* segment or index scan   */
	BYTE_INT 	clustered;
	/* if an index scan, specifies nature of index */
	INDEX_DESC 	scan_attr;	/* attribute used for index scan */
	ANYVALUE 	range_value[2];	/* operands to range operator*/ 
	int			range_operator[2];
	int		select_size;	
	BYTE_INT	select_count;	/* # of select clauses in 
					QTNODE_PARMS	*/

} SREPLACE_DESC;


/*
** This descriptor contains the information required by the
** insert operator. Put in by Anoop.
*/

typedef struct {
	NAME 		relname;	/* relation to be updated */
	int		tuple_length;	/* bytes in a tuple */
	int		index_cnt;	/* number of secondary indices to be 
					updated.  Index information is 
					contained in QT_NODE_PARAMS  */

	int		in_streams;  	/*Number of nodes the sreplace
					    query is sent. Used by the insert
					    process to expect these many
					    end-of-files. */
} INSERT_DESC;

/*
** This is a descriptor for the join operator.
*/
typedef struct {
	short 		no_of_clauses;
	short 		numattr_left, numattr_right;
	short		build_tuplen;	/* length of tuple, build rel	*/
	short		probe_tuplen;	/* length of tuple, probe rel	*/
	BYTE_INT	overflow_disk;	/* name of overflow disk 	*/ 
	BYTE_INT	build_streams;	/* count(build input streams)	*/
	BYTE_INT	probe_streams;	/* count(probe input streams)	*/
} JOIN_DESC;

/*
** Scalar aggregate descriptor
*/
typedef struct {
	BYTE_INT	aggregate_op;	/* type of aggregate operator	*/
	BYTE_INT	in_streams;	/* count(input streams)		*/
	BYTE_INT	num_nodes;	/* number of nodes this operator 
					   is activated in 		*/ 
	short		tuplen;		/* tuple length in bytes	*/
	ATTR_INFO	aggr_attr;	/* target attribute		*/
	int		patch_position;
	BYTE_INT	aggrsch_node;
	BYTE_INT	aggrsch_part;
	BYTE_INT	aggrsch_port;
} SCALAGGR_DESC;

/*
** Aggregate function descriptor
*/
typedef struct {
	BYTE_INT	aggregate_op;	/* type of aggregate operator	*/
	BYTE_INT	in_streams;	/* count(input streams)		*/
	short		tuplen;		/* tuple length in bytes	*/
	BYTE_INT	nojoin;
	BYTE_INT	proj_count;
	ATTR_INFO	aggr_attr;	/* target attribute		*/
	ATTR_INFO	by_attr;	/* by clause attributes		*/
} AGGR_FUNC_DESC;


/*
** Duplicate elimination descriptor
*/
typedef struct {
	BYTE_INT	in_streams;	/* count(input streams)		*/
	BYTE_INT	project_count;	/* attributes to be permuted	*/
	short		tuplen;		/* tuple length in bytes	*/
} DUPLICATE_ELIM_DESC;


/*
** Retrieve into descriptor
*/
typedef struct {
	NAME		relname;	/* result relation name		*/
	int		volid;
	int		numpages;
	int		extentfillfactor,pagefillfactor;
	BYTE_INT	in_streams;	/* count(input streams)		*/
	BYTE_INT	sort_output;	/* boolean; call wiss_sort	*/
	ATTR_INFO	sort_attr;
} STORE_DESC;


/*
** Collect and return result tuples to host via spooler
*/
typedef struct {
	BYTE_INT	in_streams;	/* number of input streams	*/
	BYTE_INT	dest_node;	/* destination address		*/
	BYTE_INT	dest_port;	
} SPOOL_DESC;

/*
** Delay a partition of tuples until needed
*/
typedef struct {
	BYTE_INT	in_streams;	/* number of input streams	*/
	BYTE_INT	sort_output;	/* boolean; call wiss_sort	*/
	ATTR_INFO	sort_attr;
	int		tuplen;		/* length of tuples		*/
} DELAY_DESC;

/*
** Delay a partition of tuples until needed (for transitive closure)
*/
typedef struct {
	BYTE_INT	in_streams;	/* number of input streams	*/
	int		offset;		/* offset to split tuples out	*/
	int		tuplen;		/* length of tuples		*/
} TC_DELAY_DESC;

/*
** Merge sorted files locally
*/
typedef struct {
	BYTE_INT	in_streams;	/* number of input stream	*/
	ATTR_INFO	sort_attr;	/* comparison key		*/
	NAME		filename [1];	/* the last one is the output	*/
} MERGE_DESC;

/*
** Sort a file globally
*/
typedef struct {
	int		volid;		/* for both files */
	NAME		src_file;		/*	*/
	NAME		dst_file;		/*	*/
	ATTR_INFO	sort_attr;		/*	*/
	int 		tuple_length;		/* bytes in a tuple */
	BYTE_INT	sites_involved;		/*	*/
	BYTE_INT	sort_method;		/* will be used later */
} SORT_DESC;

/*
** This is a descriptor for the create relation operation.
*/
typedef struct {
	int		volid;
	NAME		filename;
	int		numpages;
	int		extentfillfactor,pagefillfactor;
} C_REL_DESC;

/*
** This is a descriptor for the destroy relation operation.
*/
typedef struct {
	int	volid;
	NAME	filename;
} D_REL_DESC;

/*
** This is a descriptor for the create index operation.
*/
typedef struct {
	int		volid,index_num;
	BYTE_INT	primary,sorted;
	NAME		filename;
	KEYINFO 	keyattr;
	int		fillfactor;
} C_IDX_DESC;

/*
** This is a descriptor for the destroy index operation.
*/
typedef struct {
	int		volid,index_num;
	NAME		filename;
} D_IDX_DESC;

/*
** Header for QUERY_PKT packets
*/
typedef struct {
	enum QPKT_TYPE	packet_type;	/* distinguish the packet	*/
	BYTE_INT	qpkt_count;	/* number of qpacket pieces	*/
	BYTE_INT	qm_host;	/* query manager host address	*/
	BYTE_INT	qm_port;	/* query manager port name	*/
	BYTE_INT	filtered;	/* use filtering?		*/
	short		join_nodes;	/* #nodes for join, for filters */
	short		query_id;	/* increasing counter value	*/
	int 		TRANS_ID;	/* Transaction id               */
	int		pkt_length;	/* total length of query packet */
	int		rootnode;	/* offset to root of query tree	*/
} QPKT_HEADER;

/* 
** This structure is needed for hybrid join queries
** it is included as part of the qtnode_header.
*/
typedef struct {
        int          part_cnt;       /* number of partitioning entries       */
        int          joinnodes;      /* number of joining sites              */
        int          num_joins;      /* total number of joins to compute     */
        PORT_TABLE   partition1;     /* copy of join active_sites            */
	NETADDR	     *hybrid_parts;  /* hybrid join partitions               */
        struct qtnode_header *new_qtnode;   /* modified join descriptor      */
} HYBRID_INFO;


/*
** This is the structure of the header of QTNODE descriptors.
** This structure also serves as a header for MINI_QUERY packets.
*/
typedef struct qtnode_header {
	enum QPKT_TYPE	packet_type;	/* distinguish the packet	*/
	enum OPCODE	opcode;		/* operator type		*/
	enum SPLIT_TYPE	split_type;	/* range, hashed or random	*/
	ATTR_INFO 	split_attr[2];	/* split attribute descriptor	*/
	int		split_count;	/* # of split table entries	*/
	PORT_TABLE	source_ports;	/* identifies tuple sources	*/
	PORT_TABLE	active_sites;	/* node/ports of qtnode tasks	*/
	BYTE_INT	process_count;	/* number of query node tasks	*/
	BYTE_INT	filtered;	/* apply bit vector filtering?	*/
	int		filtersize;	/* BITS in bit vector filter	*/
	short		initial_memory[MAXNODE];/* initial memory grants*/
	int		required_memory;/* required memory allocation	*/
	int		stream_cnt;	/* expected #of end-of-streams	*/
	int		qtnode_size;	/* bytes in QTNODE descriptor	*/ 
	int 		TRANS_ID;	/*Transaction id*/
	BYTE_INT	activate_trans; /* flag, activate the transaction? */
	BYTE_INT	local_trans;	/* is this op (transaction) local? */
	int		qry_count;	/* # of queries - used by aggr  */
	BYTE_INT	hybrid_join;	/* is this a hybrid join?	*/
	HYBRID_INFO	hybrid_info;   /* hybrid join information	*/
	int 		left;		/* left branch -- byte offset	*/
	int 		right;		/* right branch -- byte offset	*/
} QTNODE_HEADER;

