/*
 * 	display.c
 *	
 *	Handles all drawings and widgets instantiation.
 *
 *	1. display_clear(void)
 *	   
 *	   Clean the display and get ready for new request.
 *
 *	2. display_flush(void)
 *
 *	   Actually popup the picture.
 *
 *	3. display_fill_status()
 *
 *	   
 *	4. display_add_join()
 *	5. display_add_select()
 *	6. display_add_store()
 *	7. display_add_spool()
 *
 */

#define XAW_BC

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Paned.h>

#include <X11/Xaw/Form.h>
#include "xgamma.h"
#include "MyWidget/Join.h"
#include "MyWidget/Select.h"
#include "MyWidget/Store.h"
#include "MyWidget/Spool.h"
#include "display.h"


#ifdef old
static Widget shell = 0;	 /* pop up shell for query packet */
static Widget top_pane = 0;	 /* contains tree_viewport and status window */
static Widget status_label = 0;	 /* label showing status */
static Widget tree_viewport = 0; /* contains tree_form */
static Widget tree_form = 0;	 /* contains all the nodes and edges */

static Widget stacktop = 0;	 /* points to last node inserted */
#endif old
Widget shell = 0;	 /* pop up shell for query packet */
Widget top_pane = 0;	 /* contains tree_viewport and status window */
Widget status_label = 0;	 /* label showing status */
Widget tree_viewport = 0; /* contains tree_form */
Widget tree_form = 0;	 /* contains all the nodes and edges */

Widget stacktop = 0;	 /* points to last node inserted */

static Arg g_args[] = {
    { XtNfromVert, 0 },
    { XtNdefaultDistance, 0 },
    { XtNborderWidth, 0 },
};


static char namebuf[10];	/* for widget naming */
static int node_index;		/* for unique naming of widget */

static int display_width;	/* width of display */
static int display_height;	/* height of display */


static char status_buf[80];

static void display_setup();
static void display_create_status_label();


/*
 * 	display_clear
 *
 *	Clears the display and call display_setup to create a new
 *	instance of shell for the next display request.
 */
void display_clear()
{
    static int inited = 0;
    
    if (shell)  {
	/*
	 * Destroy and recreate later
	 */
	XtDestroyWidget(shell);
    }
    
    display_setup(toplevel);	/* recreate everything */
    /*
     * reset stack and node index
     */
    stacktop = 0;
    node_index = 0;
    
    if (!inited) {
	/*
	 * Get the width and height of the physical display
	 */
	XWindowAttributes windowattr;
	inited++;
	XGetWindowAttributes(XtDisplay(toplevel), 
			     RootWindow(XtDisplay(toplevel), 
					*((int*)XtScreen(toplevel))),
			     &windowattr);
	display_width = windowattr.width;
	display_height = windowattr.height;
    }
}


#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>
struct _CorePart *MyCorePtr;

/*
 * 	display_flush()
 *
 *	Put the picture created on the screen.
 */
void display_flush()
{
    
    if (shell)  {
	/*
	 * Pop it up
	 */
	XtPopup(shell, XtGrabNonexclusive);
    }
}


/*
 *	display_fill_status()
 *
 * 	Fill up status_buf with appropriate status + comment
 */
void display_fill_status(op, proc, stat, comment)
int op, proc, stat;	/* operator, processor, status */
char* comment;		/* additional comment */
{
    char* opstr;
    char* statstr;
    char procstr[3];
    static Arg args[] = {
	{ XtNlabel, (XtArgVal) status_buf },
    };
    extern char* op_str();
    extern char* status_str();
    
#ifdef DEBUG
    printf("Op: %d  Proc: %d  Stat: %d  Comment: %s\n", op, proc, stat, comment);
#endif DEBUG
    
    opstr = (op >= 0) ? op_str(op) : NULL;
    statstr = (op >= 0) ? status_str(stat) : NULL;
    if (proc >= 0)  sprintf(procstr, "%d", proc); else procstr[0] = '\0';
    
    sprintf(status_buf, "%s: Processor %s %s; %s", opstr, 
	    procstr, statstr, comment);
    
    XtSetValues(status_label, args, (int) XtNumber(args));
}


/*
 *	display_setup()
 * 
 *	Set up shell, top_pane, status_label, tree_viewport, and 
 *	tree_form widgets.
 *
 *		shell is parent of top_pane.
 *		top_pane is parent of tree_viewport and status_label.
 *		tree_viewport is parent of tree_form.
 *
 *	All nodes will be children of tree_form.
 *	
 */
static void display_setup(parent)
Widget parent;
{
    Arg arg; 
    
    /*
     * Create Pop up Shell
     */
    shell = XtCreatePopupShell("xgamma-shell", topLevelShellWidgetClass,
			       parent, NULL, 0);
    
    /*
     * Create top pane
     */
    top_pane = XtCreateManagedWidget("xgamma-pane", panedWidgetClass,
				     shell, NULL, 0);
    
    /*
     * Create status_label along with start up message
     */
    strcpy(status_buf, "QUERY TREE GENERATED");
    XtSetArg(arg, XtNlabel, status_buf);
    status_label = XtCreateManagedWidget("xgamma-status", labelWidgetClass,
					 top_pane, &arg, 1);
    XtSetArg(arg, XtNheight, 0);
    XtGetValues(status_label, &arg, 1);
    arg.value += 50;
    XtSetValues(status_label, &arg, 1);
    XtSetArg(arg, XtNjustify, XtJustifyLeft);
    XtSetValues(status_label, &arg, 1);
    
    /*
     * Create tree_viewport
     */
    tree_viewport = XtCreateManagedWidget("xgamma-vp", viewportWidgetClass,
					  top_pane, NULL, 0);
    
    XtSetArg(arg, XtNdefaultDistance, (XtArgVal) 0);
    
    /* 
     * Create tree_form
     */
    tree_form = XtCreateManagedWidget("xgamma-form", formWidgetClass, 
				      tree_viewport, &arg, 1);
    
    XtSetArg(arg, XtNallowVert, (XtArgVal) True);
    XtSetValues(tree_viewport, &arg, 1);
    XtSetArg(arg, XtNallowHoriz, (XtArgVal) True);
    XtSetValues(tree_viewport, &arg, 1);
    
    node_index = 0;
}


/*
 * 	display_add_spool()
 *
 *	Add a spool node to tree_form
 */
Widget display_add_spool()
{
    Widget w;
    
    g_args[0].value = (XtArgVal) stacktop;
    sprintf(namebuf, "%d", node_index);
    w = XtCreateManagedWidget(namebuf, spoolWidgetClass, tree_form, 
			      g_args, (int) XtNumber(g_args));
    stacktop = w;
    node_index++;
    
    return w;
}


/*
 *	display_add_store()
 *
 *	Add a store node to tree_form
 */
Widget display_add_store()
{
    Widget w;
    
    g_args[0].value = (XtArgVal) stacktop;
    sprintf(namebuf, "%d", node_index);
    w = XtCreateManagedWidget(namebuf, storeWidgetClass, tree_form, 
			      g_args, (int) XtNumber(g_args));
    stacktop = w;
    node_index++;
    
    return w;
}


/*
 *	display_add_join()
 *
 *	Add a join node to tree_form
 */
Widget display_add_join()
{
    Widget w;
    
    g_args[0].value = (XtArgVal) stacktop;
    sprintf(namebuf, "%d", node_index);
    w = XtCreateManagedWidget(namebuf, joinWidgetClass, tree_form, 
			      g_args, (int) XtNumber(g_args));
    stacktop = w;
    node_index++;
    
    return w;
}


/*
 *	display_add_select()
 *
 *	Add a select node to tree_form
 */
Widget display_add_select()
{
    Widget w;
    
    g_args[0].value = (XtArgVal) stacktop;
    sprintf(namebuf, "%d", node_index);
    w = XtCreateManagedWidget(namebuf, selectWidgetClass, tree_form, 
			      g_args, (int) XtNumber(g_args));
    stacktop = w;
    node_index++;
    
    return w;
}
