#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h> 
#include <netdb.h> 
#include "wiss.h"
#include "qpacket.h"
#include "gpkt.h"
    
    
static void g_extract();
static void g_flush();
static void g_gen();
static void g_putbuf();
static void g_send();
static void g_getstat();

#define MAXOP	30

static u_char buffer[MAXOP];
static int bufindex;
static int inited = 0;
static int sock;

extern void bcopy();
extern void bzero();
extern int socket();


void main(argc, argv)
int argc;
char** argv;
{
    int fd;
    char* fname;
    char inbuf[10000];
    
    if (argc != 2)  {
	fprintf(stderr, "Usage: gpkt filename\n");
	exit(1);
    }
    
    fname = argv[1];
    
    if ((fd = open(fname, O_RDONLY)) < 0)  {
	fprintf(stderr, "Fatal: cannot open file %s\n", fname);
	exit(1);
    }
    
    if (read(fd, inbuf, 10000) < 0)  {
	fprintf(stderr, "Fatal: cannot read file %s\n", fname);
	exit(1);
    }
    /*
      g_init("oberon");
      g_q2gpkt(inbuf);
      
      getchar();
      */
    g_getstat("update");
    
    close(sock);
}


static void g_getstat(fname)
char* fname;
{
    FILE* f;
    
    f = freopen(fname, "r", stdin);
    if (!f)  {
    	fprintf(stderr, "Cannot open file %s\n", fname);
	exit(1);
    }
    
    printf("Calling parser now\n");
    yyparse();
}


void g_init(server)
char* server;
{
    struct sockaddr_in saddr;	/* server's address */
    struct sockaddr_in myaddr;
    struct hostent* hostp;
    char* routine = "g_init()";
    
    if (inited)  {
	(void) fprintf(stderr, "%s: called more than once\n", routine);
	return;
    }
    inited = 1;
    
    
    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0)  {
	(void) fprintf(stderr, "%s: cannot create socket\n", routine);
	inited = 0;
	return;
    }
    
    bzero((char*) &myaddr, sizeof(myaddr));
    myaddr.sin_addr.s_addr = INADDR_ANY;
    myaddr.sin_family = AF_INET;
    
    if (bind(sock, &myaddr, sizeof(myaddr)) < 0)   {
	(void) fprintf(stderr, "%s: cannot bind socket\n", routine);
	inited = 0;
	close(sock);
	return;
    }
    
    
    hostp = gethostbyname(server);
    if (hostp == NULL)  {
	(void) fprintf(stderr, "%s: cannot get host entity\n", routine);
	inited = 0;
	close(sock);
	return;
    }
    
    bcopy((char*) hostp->h_addr, (char*) &(saddr.sin_addr), 
	  hostp->h_length);
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons((u_short) X_Port);
    
    printf("connecting to %s\n", hostp->h_name);
    
    if (connect(sock, &saddr, sizeof(saddr)) < 0)  {
	(void) fprintf(stderr, "%s: connection attempt failed\n", routine);
	perror("connect");
	inited = 0;
	close(sock);
	return;
    }
}


void g_q2gpkt(packet)
char* packet;
{
    bufindex = 0;
    g_extract(((QPKT_HEADER*) packet)->rootnode, packet);
    g_flush();
}


static void g_extract(root, packet)
int root;
char* packet;
{
    char* ptr;
    
    if (root)  {
	ptr = &(packet[root]);
	g_gen(ptr);
	g_extract(((QTNODE_HEADER*) ptr)->right, packet);
	g_extract(((QTNODE_HEADER*) ptr)->left, packet);
    }
}


static void g_gen(ptr)
char* ptr;
{
    char* pos = ptr;
    QTNODE_HEADER* nhdr = (QTNODE_HEADER*) ptr;
    enum OPCODE qry_op = nhdr->opcode;
    
    pos += sizeof(QTNODE_HEADER);
    
    switch(qry_op)  {
    case SPOOL:
	printf("got a SPOOL node\n");
	g_putbuf(GGSPOOL);
	break;
    case STORE:
	printf("got a STORE node\n");
	g_putbuf(GGSTORE);
	break;
    case JOIN:
	printf("got a JOIN node\n");
	g_putbuf(GGJOIN);
	break;
    case SELECT:
	printf("got a SELECT node\n");
	g_putbuf(GGSELECT);
	break;
    default:
	printf("g_gen(): Unknown node\n");
	break;
    }
}


static void g_putbuf(opcode)
int opcode;
{
    char* routine = "g_putbuf()";
    if (bufindex < MAXOP - 1)  
        buffer[bufindex++] = (u_char) opcode;
    else
	printf("%s: buffer overflowed\n", routine);
}


static void g_flush()
{
    GPKT_HDR hdr;
    int i;
    
    hdr.magic = htons((u_short) GMAGIC);
    hdr.type = htons((u_short) GGTREE);
    hdr.nodecnt = htons((u_short) bufindex);
    
    printf("Sending a packet of %d nodes\n", bufindex);
    
    g_send((char*) &hdr, GPKT_HDR_LEN);
    
    g_send((char*) buffer, bufindex);
}


static void g_send(buf, len)
char* buf;
int len;
{
    int sentcnt = 0;
    int i;
    char* routine = "g_send()";
    
    while (sentcnt < len)  {
	int c = send(sock, &(buf[sentcnt]), len - sentcnt, 0);
	if (c < 0)  {
	    fprintf(stderr, "%s: send() failed\n", routine);
	    perror("send");
	    exit(1);
	}
	sentcnt += c;
    }
    
    printf("Sent %d bytes\n\t", sentcnt);
    for (i = 0; i < sentcnt; i++)  {
	printf("%x ", buf[i]);
    }
    printf("\n");
}


void g_update(stat, proc, op)
int stat, proc, op;
{
    GPKT_HDR hdr;
    GPKT_UPD upd;

    hdr.magic = GMAGIC;
    hdr.type = GGUPD;
    hdr.magic = htons(hdr.magic);
    hdr.type = htons(hdr.type);

    upd.processor = proc;
    upd.op = op;
    upd.stat = stat;

    /* print_update(&upd);*/

    g_send((char*) &hdr, GPKT_HDR_LEN);
    g_send((char*) &upd, GPKT_UPD_LEN);

    sleep(1);

}
