#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <X11/Intrinsic.h>
#include "xgamma.h"
#include "tree.h"
#include "gpkt.h"
#include "input.h"
    
    
static void input_accept();
static void input_recv();
static void input_process();

static XtAppContext context;

extern char* malloc();

/*
 * 	Initializes the input socket stream and setup interrupt mechanism to
 * 	call input_accept() to accept a connection.
 */
void input_init(cntxt)
XtAppContext cntxt;
{
    char* routine = "input_init()";
    struct sockaddr_in addr;
    int listen_sock;
    
    context = cntxt;
    
    listen_sock = socket(PF_INET, SOCK_STREAM, 0);
    if (listen_sock < 0)  {
	fprintf(stderr, "%s: cannot create socket\n", routine);
	perror("socket");
	exit(1);
    }
    
    addr.sin_family = AF_INET;
    addr.sin_port = htons((u_short) X_Port);
    addr.sin_addr.s_addr = INADDR_ANY;
    if (bind(listen_sock, (struct sockaddr*) &addr, sizeof(addr)) < 0)  {
	fprintf(stderr, "%s: cannot bind\n", routine);
	perror("bind");
	exit(1);
    }
    
    if (listen(listen_sock, 5) < 0)  {
	fprintf(stderr, "%s: cannot listen\n", routine);
	perror("listen");
	exit(1);
    }
    
    XtAppAddInput(context, listen_sock, XtInputReadMask, input_accept, 0);
}


static void input_accept(data, sockp, id)
caddr_t data;
int* sockp;
XtInputId* id;
{ 
    int newsock;
    int oldsock = *sockp;	/* the old socket for listening */
    struct sockaddr_in remote;
    int remotelen;
    
#ifdef DEBUG
    printf("accepting a connection\n");
#endif
    
    if ((newsock = accept(oldsock, (struct sockaddr*) &remote, 
			  &remotelen)) < 0)  {
	perror("accept\n");
	exit(1);
    }
    
    /*
     * Accepted a connection. Close listen socket and switch input to
     * input_process()
     */
    close(oldsock);
    XtRemoveInput(*id);
    XtAppAddInput(context, newsock, XtInputReadMask, input_process, 0);
}


/*
 * Read in a packet from backend
 */
static void input_process(data, sockp, id)
caddr_t data;
int* sockp;
XtInputId* id;
{
    GPKT_HDR hdr;
    int sock = *sockp;
    char* routine = "input_process()";
    
    XtRemoveInput(*id);		/* disable interrupt */
    
#ifdef DEBUG
    printf("A packet came in\n");
#endif
    
    /*
     * read in packet header
     */
    input_recv(sock, (char*) &hdr, GPKT_HDR_LEN);
    hdr.magic = ntohs(hdr.magic);
    hdr.nodecnt = ntohs(hdr.nodecnt);
    hdr.type = ntohs(hdr.type);
    
    
#ifdef DEBUG
    printf("magic = %d\n", hdr.magic);
    printf("nodecnt = %d\n", hdr.nodecnt);
    printf("type = %d\n", hdr.type);
#endif
    
    if (hdr.magic != GMAGIC)  {
	fprintf(stderr, "%s: bad magic number\n", routine);
	exit(1);
    }
    
    if (hdr.type != GGTREE && hdr.type != GGUPD && hdr.type != GGEND)  {
	fprintf(stderr, "%s: nonstandard header type\n", routine);
	exit(1);
    }
    
    if (hdr.type == GGUPD)  {
	GPKT_UPD upd;
	
	input_recv(sock, (char*) &upd, GPKT_UPD_LEN);
#ifdef DEBUG
	print_update(&upd);
#endif
	tree_update((int) upd.processor, (int) upd.op, (int) upd.stat);
    }
    
    if (hdr.type == GGTREE)  {
	GPKT_NODE* nodebuf;
	int i;
        nodebuf = (GPKT_NODE*) malloc(hdr.nodecnt*sizeof(GPKT_NODE));
        if (nodebuf == NULL)  {
	    (void) fprintf(stderr, "%s: out of memory\n", routine);
	    exit(1);
	}
	
        input_recv(sock, (char*) nodebuf, (int) hdr.nodecnt*sizeof(GPKT_NODE));
	for (i = 0; i < hdr.nodecnt; i++) 
	    nodebuf[i].op = ntohl(nodebuf[i].op);
	
	tree_create(nodebuf, (int) hdr.nodecnt);

	free((char*) nodebuf);
    }
    
    if (hdr.type == GGEND)  {
	printf("End of transmission\n");
	exit(1);
	return;
    }
    
    XtAppAddInput(context, sock, XtInputReadMask, input_process, 0);
}


/*
 * Read 'len' bytes from sock
 */
static void input_recv(sock, buf, len)
int sock;
char* buf;
int len;
{
    int recvcnt = 0;
    int retrycnt = 0;
    char* routine = "input_recv()";
    
    while (recvcnt < len)  {
	int c = recv(sock, &(buf[recvcnt]), len-recvcnt, 0);
	if (c < 0)  {
	    (void) fprintf(stderr, "%s: inrecoverable error - recv() failed\n",
			   routine);
	    exit(1);
	}
	recvcnt += c;
	if (recvcnt < len)   {
	    printf("sleeping 1\n");
	    sleep(1);
	    if (++retrycnt > 10)  {
		fprintf(stderr, "%s: timeout\n", routine);
		exit(1);
	    }
	}
    }
}
