#include <stdio.h>
#include <sys/types.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include "MyWidget/Store.h"
#include "MyWidget/Select.h"
#include "MyWidget/Join.h"
#include "MyWidget/Spool.h"
#include "xgamma.h"
#include "gpkt.h"
#include "display.h"
#include "tree.h"
	
	
typedef struct tnode {
    int op;		/* GGSPOOL, GGSTORE, GGJOIN, GGSELECT */
    int init_count;
    int done_count;
    Widget w1;
    Widget w2;
    struct tnode* left;
    struct tnode* right;
    struct tnode* parent;
} TNode;

static TNode* root;
static TNode* curnode;


static void tree_add_store();
static void tree_add_select();
static void tree_add_join();
static void tree_add_spool();
static void tree_clear();
static void tree_printnode();
static int tree_setactive();
static TNode* tree_newnode();


extern char* malloc();


void tree_update(processor, op, status)
int processor;
int op;
int status;
{
    Arg arg1;
    Arg arg2;
    char* routine = "tree_update()";
    char tmp[30];
    
#ifdef DEBUG
    printf("TREE_UPDATE(%x): ", curnode);
    tree_printnode(curnode);
#endif
    switch(status)  {
    case GGSTART:
	XtSetArg(arg1, XtNinitiate, processor);
	break;
    case GGOP_DONE:
	XtSetArg(arg1, XtNdone, processor);
	break;
    case GGHT_DONE:
	XtSetArg(arg1, XtNdone, processor);
	XtSetArg(arg2, XtNinitiate, processor);
	break;
    default:
	fprintf(stderr, "Unknown status code\n");
	exit(1);
    }
    
    
    tmp[0] = '\0';
    if (curnode->op == GGSELECT)  {
	if (status == GGOP_DONE)  {
	    XtSetValues(curnode->w1, &arg1, 1);
	    if (++curnode->done_count == curnode->init_count)  {
#ifdef DEBUG
		printf("done_count: %d, init_count: %d\n",
		       curnode->done_count, curnode->init_count);
#endif DEBUG
		display_fill_status(op, processor, status, 
				    "Done with SELECT. GO UP! ");
		curnode = curnode->parent;
	    }
	    else 
		display_fill_status(op, processor, status, tmp);
	    return;
	}
	if (op != GGSELECT)  {
	    XtSetValues(curnode->parent->w1, &arg1, 1);
	    display_fill_status(op, processor, status, tmp);
	    return;
	}
	XtSetValues(curnode->w1, &arg1, 1);
	display_fill_status(op, processor, status, tmp);
	return;
    }
    
    if (curnode->op == GGJOIN)  {
	if (op == GGSELECT)  {          /* start up right node */
	    curnode = curnode->right;
	    XtSetValues(curnode->w1, &arg1, 1);
	    display_fill_status(op, processor, status, tmp);
	    return;
	}
	if ((op == GGJOIN && status == GGSTART) || 
	    op == GGSTORE || op == GGSPOOL)  {
	    XtSetValues(curnode->parent->w1, &arg1, 1);
	    display_fill_status(op, processor, status, tmp);
	    return;
	}
	if (status == GGHT_DONE)  {
	    XtSetValues(curnode->w1, &arg1, 1);   /* done with join build */
	    XtSetValues(curnode->w2, &arg2, 1);   /* start up join probe */
	    display_fill_status(op, processor, status, tmp);
	    return;
	}
	if (status == GGOP_DONE)  {
	    XtSetValues(curnode->w2, &arg1, 1);
	    if (++curnode->done_count == curnode->init_count)  {
		sprintf(tmp, "Done with JOIN. GO UP! ");
		display_fill_status(op, processor, status, tmp);
		curnode = curnode->parent;
	    }
	    else
		display_fill_status(op, processor, status, tmp);
	    return;
	}
    }
    
    if (curnode->op == GGSPOOL || curnode->op == GGSTORE)  {
	if (status == GGOP_DONE)  {
	    XtSetValues(curnode->w1, &arg1, 1);
	    if (++curnode->done_count == curnode->init_count)  {
		curnode = curnode->parent;
		sprintf(tmp, "QUERY COMPLETED !!!");
		display_fill_status(op, processor, status, tmp);
	    }
	    else 
		display_fill_status(op, processor, status, tmp);
	    return;
	}
	XtSetValues(curnode->w1, &arg1, 1);
	display_fill_status(op, processor, status, tmp);
	return;
    }
    
    /*
     * never reached if no error
     */
    fprintf(stderr, "%s: illegal update\n", routine);
    exit(1);
}

void tree_create(nodebuf, count)
GPKT_NODE nodebuf[];
int count;
{
    int i;
    char* routine = "tree_create()";
    
    tree_clear(root);
    root = NULL;
    curnode = NULL;
    display_clear();
    for (i = 0; i < count; i++)  {
	switch(nodebuf[i].op)  {
	case GGSELECT:
	    tree_add_select(nodebuf[i].bmap);
#ifdef DEBUG
	    printf("Adding select node(curnode=%x,root=%x)\n", curnode, root);
#endif DEBUG
	    break;
	case GGJOIN:
	    tree_add_join(nodebuf[i].bmap, nodebuf[i+1].bmap);
#ifdef DEBUG
	    printf("Adding join node(curnode=%x,root=%x)\n", curnode, root);
#endif DEBUG
	    i++;	/* skip right node */
	    break;
	case GGSPOOL:
	    tree_add_spool(nodebuf[i].bmap);
#ifdef DEBUG
	    printf("Adding spool node(curnode=%x,root=%x)\n", curnode, root);
#endif DEBUG
	    break;
	case GGSTORE:
	    tree_add_store(nodebuf[i].bmap);
#ifdef DEBUG
	    printf("Adding store node(curnode=%x,root=%x)\n", curnode, root);
#endif DEBUG
	    break;
	default:
	    fprintf(stderr, "%s: unknown node type\n", routine);
	    break;
	}
    }
    display_flush();
}

static void tree_add_store(bmap)
u_char bmap[4];
{
    Widget w;
    Widget child;
    TNode* t;
    
    
    w = display_add_store();
    child = XtSGetStoreWidget(w);
    
    t = tree_newnode(GGSTORE, child, (Widget) NULL, curnode);
    t->init_count = tree_setactive(child, bmap);
    
    if (!curnode)  
	root = t;
    else  
	curnode->left = t;
    curnode = t;
}

static void tree_add_spool(bmap)
u_char bmap[4];
{
    Widget w;
    Widget child;
    TNode* t;
    
    
    w = display_add_spool();
    child = XtSGetSpoolWidget(w);
    
    
    t = tree_newnode(GGSPOOL, child, (Widget) NULL, curnode);
    t->init_count = tree_setactive(child, bmap);
    
    if (!curnode)  
	root = t;
    else  
	curnode->left = t;
    curnode = t;
}



static void tree_add_select(bmap)
u_char bmap[4];
{
    Widget w;
    Widget child;
    TNode* t;
    
    
    w = display_add_select();
    child = XtSGetSelectWidget(w);
    
    t = tree_newnode(GGSELECT, child, (Widget) NULL, curnode);
    t->init_count = tree_setactive(child, bmap);
    
    if (!curnode)  
	root = t;
    else  
	curnode->left = t;
    curnode = t;
}


static void tree_add_join(bmap1, bmap2)
u_char bmap1[4];
u_char bmap2[4];
{
    Widget w;
    Widget child1, child2;
    Widget s;
    TNode* t;
    
    w = display_add_join();			/* Create a join widget */
    child1 = XtJGetJoinBuildWidget(w);		/* Get joinBuild widget id */
    child2 = XtJGetJoinProbeWidget(w);		/* Get joinProbe widget id */
    s = XtJGetSelectWidget(w);			/* Get select widget id */
    
    
    t = tree_newnode(GGJOIN, child1, child2, curnode);
    t->right = tree_newnode(GGSELECT, s, (Widget) NULL, t);
    t->left = NULL;
    
    t->init_count = tree_setactive(child1, bmap1);
    (void) tree_setactive(child2, bmap1);   /* init_count should be the same */
    t->right->init_count = tree_setactive(s, bmap2);
    
    if (!curnode)  {
	fprintf(stderr, "Join node cannot be top node\n");
	exit(1);
    }
    else 
	curnode->left = t;
    curnode = t;
}




static TNode* tree_newnode(type, wid1, wid2, parent)
int type;
Widget wid1, wid2;
TNode* parent;
{
    TNode* t;
    char* routine = "tree_newnode()";
    
    t = (TNode*) malloc(sizeof(TNode));
    if (t == NULL)  {
	fprintf(stderr, "%s: out of memory\n", routine);
	exit(1);
    }
    t->op = type;
    t->w1 = wid1; t->w2 = wid2;
    t->left = t->right = NULL;
    t->parent = parent;
    t->init_count = t->done_count = 0;
    
    return t;
}


static int tree_setactive(w, bmap)
Widget w;
u_char bmap[4];
{
    int i;
    int active_cnt = 0;
    static Arg arg = { XtNscheduled, 0 };
    
    
#ifdef DEBUG
    for (i = 0; i < 4; i++) printf("%x ", bmap[i]);
    printf("\n");
#endif
    
    for (i = 0; i < 32; i++) {
	if (bmap[i/8] & ((u_char) 0x80 >> (i % 8)))   {
	    arg.value = (XtArgVal) i;
#ifdef DEBUG
	    printf("site %d active\n", i);
#endif
	    XtSetValues(w, &arg, 1);
	    active_cnt++;
	}
    }
    
    return active_cnt;
}

static void tree_clear(root)
TNode* root;
{
#ifdef DEBUG
    printf("Clearing root\n");
#endif
    if (root)   {
	tree_clear(root->left);
	tree_clear(root->right);
	free((char*) root);
    }
}

static void tree_printnode(t)
TNode* t;
{
    printf("Current node is ");
    if (!t)  {
	printf("NULL\n");
	return;
    }
    switch(t->op)  {
    case GGSELECT:
	printf("SELECT");
	break;
    case GGJOIN:
	printf("JOIN");
	break;
    case GGSTORE:
	printf("STORE");
	break;
    case GGSPOOL:
	printf("SPOOL");
	break;
    default:
	fprintf(stderr, "Unknown node type\n");
	exit(1);
    }
    printf("\n");
}

