#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "EdgeP.h"
#include <stdio.h>

#define offset(field) XtOffset(EdgeWidget, field)

static XtResource resources[] = {
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel), 
	offset(edge.foreground), XtRString, "XtDefaultForeground" },
    { XtNorientation, XtCOrientation, XtRInt, sizeof(int),
	offset(edge.orientation), XtRInt, (int) NoEdge },
    };

#undef offset


static void Initialize();
static void Realize();
static void Destroy();
static void Redisplay();
static Boolean SetValues();

static void GetGC();

EdgeClassRec edgeClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"Edge",
    /* widget_size		*/	sizeof(EdgeRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	NULL,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	/* QueryGeometry */ NULL,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* edge fields */
    /* empty			*/	0
  }
};

WidgetClass edgeWidgetClass = (WidgetClass)&edgeClassRec;



static void Initialize(request, new)
Widget request, new;
{
    EdgeWidget ew = (EdgeWidget) new;
    int i;

#ifdef DEBUG
    printf("Edge::Initialize() called\n");
#endif


    GetGC(ew);

    if (ew->core.width == 0) 
	ew->core.width = 100;

    if (ew->core.height == 0)
	ew->core.height = 100;

    ew->core.border_width = 0;

}



static void Realize(w, valueMask, attrs)
Widget w;
Mask* valueMask;
XSetWindowAttributes* attrs;
{
#ifdef DEBUG
    printf("Edge::Realize() called\n");
    fflush(stdout);
#endif

    *valueMask |= CWBitGravity;
    attrs->bit_gravity = ForgetGravity;
    XtCreateWindow(w, InputOutput, (Visual *) CopyFromParent, 
						*valueMask, attrs);
}



static void Redisplay(w, event, region)
Widget w;
XEvent *event;
Region region;
{
    EdgeWidget ew = (EdgeWidget) w;
    GC gc;

#ifdef DEBUG
    printf("Edge::Redisplay() called\n");
#endif

    gc = ew->edge.gc;

    switch(ew->edge.orientation)  {
	int x1, x2, y1, y2, a1, a2;
	case NoEdge:
	    break;
	case Verticle:
	    x1 = ew->core.width/2; y1 = 0;
	    x2 = x1; y2 = ew->core.height;
	    /* Draw arrow */
	    XDrawLine(XtDisplay(w), XtWindow(w), gc, x1, y1, x2, y2);
	    a1 = x1/6; a1 = a1 > 10 ? 10 : a1;
	    a2 = y2/6; a2 = a2 > 10 ? 10 : a2;
	    XDrawLine(XtDisplay(w), XtWindow(w), gc, x1, y1, x1- a1, a2);
	    XDrawLine(XtDisplay(w), XtWindow(w), gc, x1, y1, x1+ a1, a2);
	    break;
	case Horizontal:
	    x1 = 0; y1 = ew->core.height/2;
	    x2 = ew->core.width; y2 = y1;
	    /* Draw arrow */
	    XDrawLine(XtDisplay(w), XtWindow(w), gc, x1, y1, x2, y2);
	    a1 = x2/6; a1 = a1 > 10 ? 10 : a1;
	    a2 = y1/6; a2 = a2 > 10 ? 10 : a2;
	    XDrawLine(XtDisplay(w), XtWindow(w), gc, x1, y1, a1, y1 - a2);
	    XDrawLine(XtDisplay(w), XtWindow(w), gc, x1, y1, a1, y1 + a2);
	    break;
	}

}




static Boolean SetValues(current, request, new)
Widget current, request, new;
{
    EdgeWidget curew = (EdgeWidget) current;
    EdgeWidget reqew = (EdgeWidget) request;
    EdgeWidget newew = (EdgeWidget) new;
    Boolean was_resized = False;
    Boolean redisplay = False;
    int i;

    if (curew->edge.foreground != newew->edge.foreground) {
	XtReleaseGC(new, curew->edge.gc);
	GetGC(newew);
	redisplay = True;
	}
	 
    return(redisplay );
}




static void Destroy(w)
Widget w;
{
    EdgeWidget ew = (EdgeWidget) w;

    XtDestroyGC(ew->edge.gc);
}





static void GetGC(ew)
EdgeWidget ew;
{
    XGCValues values;
    XtGCMask gc_mask = GCForeground | GCBackground 
    			| GCLineStyle | GCLineWidth ;

    values.foreground = ew->edge.foreground;
    values.background = ew->core.background_pixel;
    values.line_style = LineSolid;
    values.line_width = 0;

    ew->edge.gc = XtGetGC((Widget) ew, gc_mask, &values);
}


