#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "Node.h"
#include "Edge.h"
#include "JoinP.h"
#include <stdio.h>

#define offset(field) XtOffset(JoinWidget, join.field)
static XtResource resources[] = {
   { XtNrightNode, XtCWidget, XtRWidget, sizeof(Widget),
	offset(right_node), XtRWidget, NULL }
   };

#undef offset


static void Initialize();

#define SuperClass   ((WidgetClass) &formClassRec)

JoinClassRec joinClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) SuperClass,
    /* class_name		*/	"Join",
    /* widget_size		*/	sizeof(JoinRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	XtInheritRealize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	NULL,
    /* num_resources		*/	0,
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	XtInheritResize,
    /* expose			*/	XtInheritExpose,
    /* set_values		*/	NULL,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  {
    /* composite Class fields initialization */
    /* geometry_manager		*/	XtInheritGeometryManager,
    /* change_managed		*/	XtInheritChangeManaged,
    /* insert_child		*/	XtInheritInsertChild,
    /* delete_child		*/	XtInheritDeleteChild,
    /* extension		*/	NULL
  },
  { 
    /* constraint Class fields initialization */
    /* subresources		*/	NULL,
    /* subresource_count	*/	0,
    /* constraint_size		*/	sizeof(JoinConstraintsRec),
    /* initialize		*/	NULL,
    /* destroy			*/ 	NULL,
    /* set_values		*/	NULL,
    /* extension		*/	NULL
  },
  {
    /* form Class fields initialization */
    /* layout */ 			XtInheritLayout
  },
  { /* JoinClass fields initialization */
    /* empty			*/	0
  }
};

WidgetClass joinWidgetClass = (WidgetClass)&joinClassRec;


static void Initialize(request, new)
Widget request;
Widget new;
{
    Arg arg;
    JoinWidget jbw = (JoinWidget) new;
    static Arg NodeArgs[6] = {
	{ XtNnodeType, (XtArgVal) Unknown },
	{ XtNfromVert, (XtArgVal) NULL },
	{ XtNfromHoriz, (XtArgVal) NULL },
	{ XtNvertDistance, (XtArgVal) 0 },
	{ XtNhorizDistance, (XtArgVal) 10 },
	};
    static Arg EdgeArgs[9] = {
	{ XtNorientation, (XtArgVal) Verticle },
	{ XtNborderWidth, (XtArgVal) 0 },
	{ XtNfromVert, (XtArgVal) NULL },
	{ XtNfromHoriz, (XtArgVal) NULL },
	{ XtNvertDistance, (XtArgVal) 0 },
	{ XtNhorizDistance, (XtArgVal) 10 },
	{ XtNwidth, (XtArgVal) 0 },
	{ XtNheight, (XtArgVal) 0 },
	};
    Arg node_args[6];
    Arg edge_args[9];
    int width = 0;
    int height = 0;
    Widget children[8], *childP = children;
    int i;

    for (i = 0; i < 6; i++)  {
	node_args[i].name = NodeArgs[i].name;
	node_args[i].value = NodeArgs[i].value;
	}

    for (i = 0; i < 9; i++)  {
	edge_args[i].name = EdgeArgs[i].name;
	edge_args[i].value = EdgeArgs[i].value;
	}
    /*
     * Create Edge joining join probe with whatever that is on top
     */
    edge_args[2].value = (XtArgVal) NULL;	/* XtNfromVert */
    jbw->join.top_edge = XtCreateWidget("to", edgeWidgetClass, new, 
					edge_args, XtNumber(edge_args));
    *childP++ = jbw->join.top_edge;

    /*
     * Create JoinProbe Widget
     */
    node_args[0].value = (XtArgVal) JoinProbe;
    node_args[1].value = (XtArgVal) jbw->join.top_edge;
    jbw->join.jp = XtCreateWidget("j-p", nodeWidgetClass, new,
					node_args, XtNumber(node_args));
    *childP++ = jbw->join.jp;

    /*
     * Get width and height of JoinProbe to standardize size
     */
    width = jbw->join.jp->core.width;
    height = jbw->join.jp->core.height;

    edge_args[6].value = width;	/* XtNwidth */
    edge_args[7].value = height/2;	/* XtNheight */
    XtSetValues(jbw->join.top_edge, edge_args, XtNumber(edge_args));

    /*
     * Create Edge joining join probe with Select Widget at right
     */
    edge_args[0].value = (int) Horizontal;	/* XtNorientation */
    edge_args[2].value = (XtArgVal) jbw->join.top_edge;/* XtNfromVert */
    edge_args[3].value = (XtArgVal) jbw->join.jp;   /* fromHoriz */
    edge_args[4].value = (XtArgVal) 0;  /* XtNvertDistance */
    edge_args[5].value = (XtArgVal) 0;
    edge_args[6].value = width / 2;
    edge_args[7].value = height;
    jbw->join.right_edge = XtCreateWidget("tr", edgeWidgetClass, new,
					edge_args, XtNumber(edge_args));

    *childP++ = jbw->join.right_edge;

    /*
     * Create Select Widget at the right of JoinProbe
     */
    node_args[0].value = (int) Select;	/* XtNnodeType */
    node_args[2].value = (XtArgVal) jbw->join.right_edge;  /* fromHoriz */
    node_args[4].value = (XtArgVal) 0;
    jbw->join.right_node = XtCreateWidget("select", nodeWidgetClass,
					new, node_args, XtNumber(node_args));
    *childP++ = jbw->join.right_node;

    /*
     * Create Edge joining join probe with join build
     */
    edge_args[0].value = (int) Verticle;	/* XtNorientation */
    edge_args[2].value = (XtArgVal) jbw->join.jp;  /* XtNfromVert */
    edge_args[3].value = (XtArgVal) 0;   /* fromHoriz */
    edge_args[5].value = (XtArgVal) 10;
    edge_args[6].value = width;	/* XtNwidth */
    edge_args[7].value = height/2;	/* XtNheight */
    jbw->join.middle_edge = XtCreateWidget("tm", edgeWidgetClass, new,
					edge_args, XtNumber(edge_args));
    *childP++ = jbw->join.middle_edge;

    /*
     * Create JoinBuild Widget
     */
    node_args[0].value = (XtArgVal) JoinBuild;
    node_args[1].value = (XtArgVal) jbw->join.middle_edge;  /* fromVert */
    node_args[2].value = (XtArgVal) 0;  /* fromHoriz */
    node_args[4].value = (XtArgVal) 10; /* distHoriz */
    jbw->join.jb = XtCreateWidget("j-b", nodeWidgetClass, new,
					node_args, XtNumber(node_args));
    *childP++ = jbw->join.jb;

    XtManageChildren(children, (Cardinal) (childP - children));
}



Widget XtJGetSelectWidget(w)
Widget w;
{
    JoinWidget jbw = (JoinWidget) w;

    if (strcmp(XtClass(w)->core_class.class_name, "Join"))  {
	XtErrorMsg(
		"XtJGetSelectWidget() called with wrong widget type\n");
	}

    return((Widget) jbw->join.right_node);
}


Widget XtJGetJoinBuildWidget(w)
Widget w;
{
    JoinWidget jbw = (JoinWidget) w;

    if (strcmp(XtClass(w)->core_class.class_name, "Join"))  {
	XtErrorMsg(
		"XtJGetJoinBuildWidget() called with wrong widget type\n");
	}

    return((Widget) jbw->join.jb);
}


Widget XtJGetJoinProbeWidget(w)
Widget w;
{
    JoinWidget jbw = (JoinWidget) w;

    if (strcmp(XtClass(w)->core_class.class_name, "Join"))  {
	XtErrorMsg(
		"XtJGetJoinProbeWidget() called with wrong widget type\n");
	}

    return((Widget) jbw->join.jp);
}
