#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "NodeP.h"
#include <stdio.h>

#define offset(field) XtOffset(NodeWidget, field)

static XtResource resources[] = {
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel), 
	offset(node.foreground), XtRString, "XtDefaultForeground" },
    { XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(node.font), XtRString, "XtDefaultFont" },
    { XtNnodeType, XtCNodeType, XtRNodeType, sizeof(NodeType),
	offset(node.node_type), XtRImmediate, (caddr_t) Unknown },
    { XtNinitiate, XtCProcessor, XtRInt, sizeof(int),
	offset(node.initiate), XtRImmediate, (caddr_t) -1 },
    { XtNscheduled, XtCProcessor, XtRInt, sizeof(int),
	offset(node.scheduled), XtRImmediate, (caddr_t) -1 },
    { XtNidle, XtCProcessor, XtRInt, sizeof(int),
	offset(node.idle), XtRImmediate, (caddr_t) -1 },
    { XtNdone, XtCProcessor, XtRInt, sizeof(int),
	offset(node.done), XtRImmediate, (caddr_t) -1 },
    { XtNinitiateColor, XtCForeground, XtRPixel, sizeof(Pixel), 
	offset(node.initiateColor), XtRString, "XtDefaultForeground" },
    { XtNdoneColor, XtCForeground, XtRPixel, sizeof(Pixel), 
	offset(node.doneColor), XtRString, "XtDefaultForeground" },
    { XtNscheduledColor, XtCForeground, XtRPixel, sizeof(Pixel), 
	offset(node.scheduledColor), XtRString, "XtDefaultForeground" },
    };

#undef offset


static void Initialize();
static void Realize();
static void Destroy();
static void Resize();
static void Redisplay();
static Boolean SetValues();
static XtGeometryResult QueryGeometry();

static void GetGC();
static void GetSmallFont();
static void SetTextWidthAndHeight();
static void DrawState();

static char* nodeNames[] = { 	"NoName",
				"Spool",
				"Store",
    				"Select",
    				"Join-Build",
    				"Join-Probe" };

NodeClassRec nodeClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"Node",
    /* widget_size		*/	sizeof(NodeRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	/* QueryGeometry */ NULL,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* node fields */
    /* empty			*/	0
  }
};

WidgetClass nodeWidgetClass = (WidgetClass)&nodeClassRec;



static void Initialize(request, new)
Widget request, new;
{
    NodeWidget nw = (NodeWidget) new;
    int i;

#ifdef DEBUG
    printf("Node::Initialize() called\n");
#endif


    GetSmallFont(nw);
    GetGC(nw);

    SetTextWidthAndHeight(nw);

    if (nw->core.width == 0) 
	nw->core.width = nw->node.name_width + 90;

    if (nw->core.height == 0)
	nw->core.height = nw->node.name_height + 40;

    for (i = 0; i < NUMPROC; i++)  {
	nw->node.processor[i] = idle; 
	}

    Resize((Widget) nw);

}



static void Realize(w, valueMask, attrs)
Widget w;
Mask* valueMask;
XSetWindowAttributes* attrs;
{
#ifdef DEBUG
    printf("Node::Realize() called\n");
    fflush(stdout);
#endif

    *valueMask |= CWBitGravity;
    attrs->bit_gravity = ForgetGravity;
    XtCreateWindow(w, InputOutput, (Visual *) CopyFromParent, 
						*valueMask, attrs);
}



static void Redisplay(w, event, region)
Widget w;
XEvent *event;
Region region;
{
    NodeWidget nw = (NodeWidget) w;
    GC gc;
    int x;
    int i;
    char *node_name;

#ifdef DEBUG
    printf("Node::Redisplay() called\n");
#endif


    gc = nw->node.gc;
    if (nw->node.too_small)  {
	XDrawString(XtDisplay(w), XtWindow(w), gc, 0, 
			nw->node.font->max_bounds.ascent, 
			"Too Small", 9);
	return;
	}
    node_name = nodeNames[ (int) nw->node.node_type];
    XDrawString(XtDisplay(w), XtWindow(w), gc, nw->node.name_x,
			nw->node.name_y + nw->node.font->max_bounds.ascent,
			node_name, strlen(node_name));
    
    x = 0;
    for (i = 0; i < NUMPROC; i++)  {
	int width = nw->node.proc_width;

	if (i < (nw->node.scale_factor)) width++;
	if (i%4 == 0)  {
	    char buf[3];
	    sprintf(buf, "%d", i);
	    XDrawString(XtDisplay(w), XtWindow(w), nw->node.small_gc, x, 
			nw->node.name_y+nw->node.name_height+
			nw->node.small_font->max_bounds.ascent,
			buf, strlen(buf));
	    }

	DrawState(nw, i, x, nw->node.first_row, width);

	x += width;
	}
}



static void Resize(w)
Widget w;
{
    NodeWidget nw = (NodeWidget) w;

#ifdef DEBUG
    printf("Node::Resize() called\n");
    fflush(stdout);
#endif

    nw->node.name_x = 0;
    nw->node.name_y = 0;
    nw->node.first_row = (nw->node.name_height)+(nw->node.number_height)+2;
    nw->node.proc_height = (nw->core.height - nw->node.first_row);
    nw->node.proc_width = nw->core.width / NUMPROC;
    nw->node.scale_factor = nw->core.width % NUMPROC;
    nw->node.idle_height = 0;
    nw->node.scheduled_height = nw->node.proc_height / 3;
    nw->node.initiate_height = nw->node.scheduled_height * 2;
    nw->node.done_height = nw->node.proc_height;

    if ((nw->node.first_row < nw->node.name_height) 
		|| (nw->node.name_width > nw->core.width)) {
	nw->node.too_small = True;
#ifdef DEBUG
        printf("first_row = %d, name_height = %d\n", nw->node.first_row,
							nw->node.name_height);
	printf("name_width = %d, core_width = %d\n", nw->node.name_width,
							nw->core.width);
	printf("Too small\n");
#endif
        }
     else 
	nw->node.too_small = False;

}



static Boolean SetValues(current, request, new)
Widget current, request, new;
{
    NodeWidget curnw = (NodeWidget) current;
    NodeWidget reqnw = (NodeWidget) request;
    NodeWidget newnw = (NodeWidget) new;
    Boolean was_resized = False;
    Boolean redisplay = False;
    Boolean state_changed = False;
    unsigned long mask = 0;
    int i;


    if (curnw->node.font != newnw->node.font)  {
	SetTextWidthAndHeight(newnw);
	was_resized = True;
	}

    if (was_resized)  {
	/*
	 * Determine if resize needed, but wasn't result from physical resize.
	 * ie due to change of font, etc.
	 */
	if (curnw->core.width == reqnw->core.width) 
	    newnw->core.width = newnw->node.name_width + 2;
	if (curnw->core.height == reqnw->core.height)
	    newnw->core.height = newnw->node.name_height + 40;
	}

    if (curnw->node.foreground != newnw->node.foreground
	|| curnw->node.font->fid != newnw->node.font->fid)  {
	XtReleaseGC(new, curnw->node.gc);
	GetGC(newnw);
	redisplay = True;
	}
	 
    if (curnw->node.idle != newnw->node.idle)  {
	if (newnw->node.idle >=0 && newnw->node.idle < NUMPROC) {
	    newnw->node.processor[newnw->node.idle] = idle;
	    state_changed = True;
	    mask |= (unsigned long) 1 << newnw->node.idle;
	    }
	newnw->node.idle = -1;
	}

    if (curnw->node.initiate != newnw->node.initiate)  {
	if (newnw->node.initiate >= 0 && newnw->node.initiate < NUMPROC)  {
	    newnw->node.processor[newnw->node.initiate] = initiate;
	    state_changed = True;
	    mask |= (unsigned long) 1 << newnw->node.initiate;
	    }
	newnw->node.initiate = -1;
	}

    if (curnw->node.scheduled != newnw->node.scheduled)  {
	if (newnw->node.scheduled >= 0 && newnw->node.scheduled < NUMPROC)  {
	    newnw->node.processor[newnw->node.scheduled] = scheduled;
	    state_changed = True;
	    mask |= (unsigned long) 1 << newnw->node.scheduled;
	    }
	newnw->node.scheduled = -1;
	}

    if (curnw->node.done != newnw->node.done)  {
	if (newnw->node.done >= 0 && newnw->node.done < NUMPROC)  {
	    newnw->node.processor[newnw->node.done] = done;
	    state_changed = True;
	    mask |= (unsigned long) 1 << newnw->node.done;
	    }
	newnw->node.done = -1;
	}

    if (!was_resized && !redisplay)  {
	if (XtIsRealized((Widget) newnw) && state_changed)  {
	    int x = 0;
	    for (i = 0; i < NUMPROC; i++)  {
		int width = newnw->node.proc_width;

		if (i < (newnw->node.scale_factor)) width++;

		if (mask >> i & 1)  {
		    DrawState(newnw, i, x, newnw->node.first_row, width);
		    }

		x += width;
		}
	    }
	}
    return(was_resized || redisplay );
}




static void Destroy(w)
Widget w;
{
    NodeWidget nw = (NodeWidget) w;

    XtDestroyGC(nw->node.gc);
}



static XtGeometryResult QueryGeometry(w, intended, preferred)
Widget w;
XtWidgetGeometry* intended;
XtWidgetGeometry* preferred;
{
    NodeWidget nw = (NodeWidget) w;

    preferred->request_mode = CWWidth | CWHeight;
    preferred->width = nw->node.name_width;
    preferred->height = nw->node.name_height + nw->node.number_height + 40;

    if (((intended->request_mode & (CWWidth|CWHeight))== (CWWidth|CWHeight))
		&& (intended->width == preferred->width) 
		&& (intended->height == preferred->height)) 
	return XtGeometryYes;
    
    if (preferred->width == w->core.width &&
		preferred->height == w->core.height)
	return XtGeometryNo;

    return XtGeometryAlmost;
}





static void GetGC(nw)
NodeWidget nw;
{
    XGCValues values;
    XtGCMask gc_mask = GCForeground | GCBackground | GCFont 
    			| GCLineStyle | GCLineWidth ;

    values.foreground = nw->node.foreground;
    values.background = nw->core.background_pixel;
    values.font = nw->node.font->fid;
    values.line_style = LineSolid;
    values.line_width = 0;

    nw->node.gc = XtGetGC((Widget) nw, gc_mask, &values);

    values.font = nw->node.small_font->fid;

    nw->node.small_gc = XtGetGC((Widget) nw, gc_mask, &values);

    values.foreground = nw->node.initiateColor;
    nw->node.initiateGC = XtGetGC((Widget) nw, gc_mask, &values);

    values.foreground = nw->node.scheduledColor;
    nw->node.scheduledGC = XtGetGC((Widget) nw, gc_mask, &values);

    values.foreground = nw->node.doneColor;
    nw->node.doneGC = XtGetGC((Widget) nw, gc_mask, &values);

}



static void GetSmallFont(nw)
NodeWidget nw;
{
    nw->node.small_font = XLoadQueryFont(XtDisplay(nw), "6x10");
    if (nw->node.small_font == NULL)  {
	XtErrorMsg("Cannot open 6x10 font\n");
	}

    nw->node.number_height = nw->node.small_font->max_bounds.ascent +
				nw->node.small_font->max_bounds.descent;
}

static void SetTextWidthAndHeight(nw)
NodeWidget nw;
{
    XFontStruct *fs = nw->node.font;
    int i;

    /* Get max width */
    nw->node.name_width = 0;
    for (i = 0; i < NUM_NODETYPE; i++)  {
	int width = XTextWidth(fs, nodeNames[i], strlen(nodeNames[i]));
	if (width > nw->node.name_width) 
	    nw->node.name_width = width;
	}

    nw->node.name_height = fs->max_bounds.ascent + fs->max_bounds.descent;
}



static void DrawState(nw, proc_num, x, y, width)
NodeWidget nw;		/* node widget */
int proc_num;		/* processor number */
int x,y;		/* starting position */
int width;		/* width of processor slot */
{
    int height;
    GC gc;

    XDrawRectangle(XtDisplay(nw), XtWindow(nw), nw->node.gc, x, 
			nw->node.first_row, width, nw->node.proc_height);
    switch(nw->node.processor[proc_num])  {
	case idle:
	    height = nw->node.idle_height;
	    gc = nw->node.gc;
	    break;
	case scheduled:
	    height = nw->node.scheduled_height;
	    gc = nw->node.scheduledGC;
	    break;
	case initiate:
	    height = nw->node.initiate_height;
	    gc = nw->node.initiateGC;
	    break;
	case done:
	    height = nw->node.done_height;
	    gc = nw->node.doneGC;
	    break;
	default:
	    XtErrorMsg("Node::DrawState(): unknown state\n");
	}

    height = height <= 0 ? 0 : height -1;
    width = width <= 0 ? 0 : width -1;
    XFillRectangle(XtDisplay(nw), XtWindow(nw), gc,
			x+1, y+1, width, height);
}
